"""Job module for MQSS Client"""

from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import Dict, Optional


class JobStatus(str, Enum):
    """Status enumeration for the job status"""

    PENDING = "PENDING"
    # NOTE: We need an intermediate status before COMPLETED for the job runner to work
    WAITING = "WAITING"
    FAILED = "FAILED"
    COMPLETED = "COMPLETED"
    CANCELLED = "CANCELLED"


class JobRequest(ABC):  # pylint: disable=too-few-public-methods
    """Base class for job requests"""

    def to_json_dict(self) -> dict:
        """Convert JobRequest to JSON dictionary"""
        raise NotImplementedError("Subclasses must implement this method")


@dataclass
class CircuitJobRequest(JobRequest):
    """Class to hold job information"""

    circuits: str
    circuit_format: str
    resource_name: str
    shots: int
    no_modify: bool
    queued: bool

    def to_json_dict(self) -> dict:
        """Convert CircuitJobRequest to JSON dictionary"""
        return {
            "circuit": self.circuits,
            "circuit_format": self.circuit_format,
            "resource_name": self.resource_name,
            "shots": self.shots,
            "no_modify": self.no_modify,
            "queued": self.queued,
        }


@dataclass
class HamiltonianJobRequest(JobRequest):
    """Class to hold Hamiltonian job information"""

    resource_name: str
    interaction_str: str
    coefficients_str: str

    def to_json_dict(self) -> dict:
        """Convert HamiltonianJobRequest to JSON dictionary"""
        return {
            "resource_name": self.resource_name,
            "interaction_str": self.interaction_str,
            "coefficients_str": self.coefficients_str,
        }


@dataclass
class Result:
    """Result Class to hold counts"""

    counts: Dict[str, int]
    timestamp_submitted: datetime
    timestamp_scheduled: datetime
    timestamp_completed: Optional[datetime] = None



"""
This module provides a client for the MQP API and HPC Offload Listener.
"""

import json
import time
from datetime import datetime
from typing import Dict, Optional, Union

from .hpc_client import HPCOffloadClient
from .job import CircuitJobRequest, HamiltonianJobRequest, JobRequest, JobStatus, Result
from .resource_info import ResourceInfo
from .rest_client import RESTClient


class MQSSClient:
    """MQSS Client class for interacting with MQP REST API / HPC Offload Listener"""

    def __init__(self, token: str, base_url: str, is_hpc: bool = False):
        self.token = token
        self.base_url = base_url
        self.client: Union[HPCOffloadClient, RESTClient]
        if is_hpc:
            self.client = HPCOffloadClient(token)
        else:
            self.client = RESTClient(token, base_url)

    def get_all_resources(self) -> Dict[str, ResourceInfo]:
        """Get resource info about all resources"""

        _resources = {}
        rsp_json = self.client.get("resources")
        for name, resource_json in rsp_json.items():
            _resources[name] = ResourceInfo.from_json_dict(resource_json)

        return _resources

    def get_resource_info(self, resource_name: str) -> Optional[ResourceInfo]:
        """Get resource info about specific resource"""
        rsp_json = self.client.get(f"resources/{resource_name}")

        if not rsp_json:
            return None

        return ResourceInfo.from_json_dict(rsp_json)

    def submit_job(self, job_request: JobRequest) -> str:
        """Submit a circuit job"""
        if isinstance(job_request, CircuitJobRequest):
            rsp_json = self.client.post(
                "job",
                job_request.to_json_dict(),
            )
        elif isinstance(job_request, HamiltonianJobRequest):
            rsp_json = self.client.post(
                "hamiltonian_job",
                job_request.to_json_dict(),
            )
        else:
            raise ValueError("Invalid job request type")

        if not rsp_json:
            raise ValueError("Invalid response from server")

        if "uuid" not in rsp_json:
            raise ValueError("UUID not found in response")

        return rsp_json["uuid"]

    def cancel_job(self, uuid: str, job_type: JobRequest) -> None:
        """Cancel a job"""
        if isinstance(job_type, CircuitJobRequest):
            self.client.delete(f"job/{uuid}")
        elif isinstance(job_type, HamiltonianJobRequest):
            self.client.delete(f"hamiltonian_job/{uuid}")
        else:
            raise ValueError("Invalid job request type")

    def job_status(self, uuid: str, job_type: JobRequest) -> JobStatus:
        """Get job status"""
        if isinstance(job_type, CircuitJobRequest):
            rsp_json = self.client.get(f"job/{uuid}/status")
        elif isinstance(job_type, HamiltonianJobRequest):
            rsp_json = self.client.get(f"hamiltonian_job/{uuid}/status")
        else:
            raise ValueError("Invalid job request type")

        if not rsp_json:
            raise ValueError("Invalid response from server")
        if "status" not in rsp_json:
            raise ValueError("Status not found in response")

        return JobStatus(rsp_json["status"])

    def job_result(self, uuid: str, job_type: JobRequest) -> Optional[Result]:
        """Get job result as JSON"""
        if isinstance(job_type, CircuitJobRequest):
            result_json = self.client.get(f"job/{uuid}/result")
        elif isinstance(job_type, HamiltonianJobRequest):
            result_json = self.client.get(f"hamiltonian_job/{uuid}/result")
        else:
            raise ValueError("Invalid job request type")

        if not result_json:
            return None

        return Result(
            counts=json.loads(result_json["result"]),
            timestamp_completed=(
                datetime.strptime(
                    result_json["timestamp_completed"], "%Y-%m-%d %H:%M:%S.%f"
                )
                if result_json["timestamp_completed"] != ""
                else None
            ),
            timestamp_submitted=datetime.strptime(
                result_json["timestamp_submitted"], "%Y-%m-%d %H:%M:%S.%f"
            ),
            timestamp_scheduled=datetime.strptime(
                result_json["timestamp_scheduled"], "%Y-%m-%d %H:%M:%S.%f"
            ),
        )

    def wait_for_job_result(self, uuid: str, job_type: JobRequest) -> Optional[Result]:
        """Wait for a job to complete and return the result"""
        timeout = 2.0
        end_status = self.job_status(uuid, job_type)
        while end_status in (JobStatus.PENDING, JobStatus.WAITING):
            time.sleep(timeout)
            end_status = self.job_status(uuid, job_type)

        if end_status == JobStatus.COMPLETED:
            return self.job_result(uuid, job_type)
        if end_status == JobStatus.FAILED:
            raise RuntimeError("Job failed")
        if end_status == JobStatus.CANCELLED:
            if isinstance(job_type, CircuitJobRequest):
                cancel_reason = self.client.get(f"job/{uuid}/cancel_reason")
            else:
                cancel_reason = self.client.get(f"hamiltonian_job/{uuid}/cancel_reason")
            raise RuntimeError(f"Job cancelled: {cancel_reason}")

        raise RuntimeError("Unknown status")

    def get_num_pending_jobs(self, resource_name: str) -> int:
        """Get the number of pending jobs for a resource"""
        rsp_json = self.client.get(f"resources/{resource_name}/num_pending_jobs")

        if not rsp_json or "num_pending_jobs" not in rsp_json:
            raise ValueError("Invalid response from server")

        return rsp_json["num_pending_jobs"]



"""Client for HPCQC Offloading
It needs to interact with the following API endpoints:
- `resources`: Get information about all resources.
- `resources/{resource_name}`: Get information about a specific resource.
- `resources/{resource_name}/num_pending_jobs`: Get the number of pending jobs for a resource.
- `job`: Submit a job.
- `job/{uuid}/status`: Get the status of a job.
- `job/{uuid}/result`: Get the result of a job.
- `job/{uuid}/cancel_reason`: Get the cancel reason of a job.
- `hamiltonian_job`: Submit a Hamiltonian job.
- `hamiltonian_job/{uuid}/status`: Get the status of a Hamiltonian job.
- `hamiltonian_job/{uuid}/result`: Get the result of a Hamiltonian job.
- `hamiltonian_job/{uuid}/cancel_reason`: Get the cancel reason of a Hamiltonian job.
The client can be initialized with a token and a base URL.
It supports both REST and HPC Offload modes.
"""

import json
import socket
import uuid
from dataclasses import dataclass

from .base_client import BaseClient
from .comm.rmq_client import RabbitMQClient

HOSTNAME = socket.gethostname().replace(" ", "_")


class HPCOffloadClient(BaseClient):
    """HPC Client for HPC Offload Listener"""

    def __init__(
        self,
        token: str,
        offload_listener_queue_name: str = f"qd_qrequest_reception_queue_{HOSTNAME}",
    ) -> None:
        """Initialize the HPC Offload Client"""
        self.token = token
        self.offload_listener_queue_name = offload_listener_queue_name
        unique_id = str(uuid.uuid4())[:8]
        self.response_queue_name = f"response_queue_{HOSTNAME}_{unique_id}"

    def get(self, path: str) -> dict:
        """GET request to the specified path"""
        # Implement GET request logic here
        with RabbitMQClient(self.token) as client:
            request = OffloadRequest(
                authorization=self.token,
                method="GET",
                request=path,
                data={},
                response_queue=self.response_queue_name,
            )
            client.declare_queue(self.response_queue_name)
            client.send(request.request_str(), self.offload_listener_queue_name)
            response = client.receive(request.response_queue)
            client.delete_queue(self.response_queue_name)
            if response is not None:
                return json.loads(response)
        return {}

    def post(self, path: str, data: dict) -> dict:
        """POST request to the specified path"""
        # Implement POST request logic here
        with RabbitMQClient(self.token) as client:
            request = OffloadRequest(
                authorization=self.token,
                method="POST",
                request=path,
                data=data,
                response_queue=self.response_queue_name,
            )
            client.declare_queue(self.response_queue_name)
            client.send(request.request_str(), self.offload_listener_queue_name)
            response = client.receive(request.response_queue)
            client.delete_queue(self.response_queue_name)
            if response is not None:
                return json.loads(response)
        return {}

    def delete(self, path: str) -> None:
        """DELETE request to the specified path"""
        # Implement DELETE request logic here
        with RabbitMQClient(self.token) as client:
            request = OffloadRequest(
                authorization=self.token,
                method="DELETE",
                request=path,
                data={},
                response_queue="",
            )
            client.send(request.request_str(), self.offload_listener_queue_name)


@dataclass
class OffloadRequest:
    """Data class for offload requests"""

    authorization: str
    method: str
    request: str
    data: dict
    response_queue: str

    def request_str(self) -> str:
        """Convert the request to a JSON string"""
        return json.dumps(self.__dict__)



"""Module for ResourceInfo class"""

import ast
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple


@dataclass
class ResourceInfo:
    """Hold information about resource needed for transpilation"""

    name: str
    qubits: int
    online: bool = False
    connectivity: Optional[List[List[int]]] = None
    instructions: Optional[List[Tuple[str, Optional[Dict[Any, Any]]]]] = None

    def __eq__(self, other) -> bool:
        return isinstance(other, ResourceInfo) and self.qubits == other.qubits

    @classmethod
    def from_json_dict(cls, resource_json: dict):
        """Return ResourceInfo object from json string"""

        if "name" not in resource_json:
            raise ValueError("Resource name not found")
        _name = resource_json["name"]

        _online = False
        _connectivity = None
        _instructions = None

        if "online" in resource_json:
            _online = resource_json["online"]
        try:
            _connectivity = ast.literal_eval(resource_json["connectivity"])
        except (
            ValueError,
            TypeError,
            SyntaxError,
            MemoryError,
            RecursionError,
            KeyError,
        ):
            pass
        try:
            _instructions = ast.literal_eval(resource_json["instructions"])
        except (
            ValueError,
            TypeError,
            SyntaxError,
            MemoryError,
            RecursionError,
            KeyError,
        ):
            pass

        return cls(
            name=_name,
            qubits=resource_json["qubits"],
            online=_online,
            connectivity=_connectivity,
            instructions=_instructions,
        )



"""Base Client for underlying communication clients of the MQSS Client"""

from abc import ABC, abstractmethod


class BaseClient(ABC):
    """Abstract base class for MQSS clients"""

    @abstractmethod
    def get(self, path: str) -> dict:
        """GET request to the specified path"""
        return {}

    @abstractmethod
    def post(self, path: str, data: dict) -> dict:
        """POST request to the specified path"""
        return {}

    @abstractmethod
    def delete(self, path: str) -> None:
        """DELETE request to the specified path"""
        return None



"""RabbitMQ Client for HPCQC"""

from dataclasses import dataclass
from json import load
from typing import Any, Dict, Optional
from warnings import warn

import pika.exceptions as pikaExceptions  # type: ignore
from pika import ConnectionParameters  # type: ignore
from pika import BlockingConnection, PlainCredentials
from pika.exceptions import AMQPError  # type: ignore


class RabbitMQClient:
    """RabbitMQ Client for HPCQC"""

    def __init__(self, token: str) -> None:
        self.token = token
        self.connection = None
        self.channel = None
        self.conn_config = ConnectionConfiguration()

    def connect(self) -> None:
        """Connect to RabbitMQ server"""
        print(f"Connecting to RabbitMQ server {self.conn_config} ...")
        self.connection = create_rmq_connection(self.conn_config)
        assert self.connection is not None
        self.channel = self.connection.channel()

    def close(self) -> None:
        """Close the connection to RabbitMQ server"""
        if self.channel:
            self.channel.close()
        if self.connection:
            self.connection.close()

    def send(self, data: str, destination: str) -> bool:
        """Send data to RabbitMQ server"""
        if not self.channel:
            return False
        try:
            self.channel.queue_declare(queue=destination, durable=True)
            self.channel.basic_publish(
                exchange="", routing_key=destination, body=data, mandatory=True
            )
        except AMQPError:
            return False
        return True

    def receive(
        self, source: str
    ) -> Optional[str]:  # Updated return type to Optional[str]
        """Receive data from RabbitMQ server"""
        if not self.channel:
            return None

        response = None
        self.channel.queue_declare(queue=source, durable=True)
        for method_frame, _, body in self.channel.consume(
            queue=source,
            inactivity_timeout=1,
            auto_ack=True,
        ):
            if body is not None:
                response = body.decode()
                print(f"Received message: {body}")
                print(f"Decoded message: {response}")
                print(f"Decoded message type: {type(response)}")
                print(f"Decoded message str: {str(response)}")

                break

        return response

    def declare_queue(self, queue_name: str) -> None:
        """Declare a queue"""
        if not self.channel:
            return
        self.channel.queue_declare(queue=queue_name, durable=True)

    def delete_queue(self, queue_name: str) -> None:
        """Delete a queue"""
        if not self.channel:
            return
        self.channel.queue_delete(queue=queue_name)

    def __enter__(self) -> "RabbitMQClient":
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()


@dataclass
class ConnectionConfiguration:
    """Configuration required for connection to RabbitMQ"""

    host: str = "localhost"
    port: int = 5672
    user: str = "guest"
    password: str = "guest"

    @classmethod
    def from_json_file(cls, config_file_path: str, conn_config_key: str = ""):
        """Read Connection Configuration from a file"""

        try:
            with open(file=config_file_path, encoding="utf-8") as json_config_file:
                config_data: Dict[str, Any] = load(json_config_file)
                config_dict: Dict[str, Any] = {}
                if conn_config_key != "" and conn_config_key in config_data:
                    config_dict = config_data[conn_config_key]
                else:
                    config_dict = config_data

                conn_config = cls(**config_dict)
        except (FileExistsError, FileNotFoundError):
            warn(f"Configuration file not found: {config_file_path}")
            return cls()

        return conn_config


def create_rmq_connection(
    conn_config: ConnectionConfiguration, raise_exceptions: bool = False
) -> Optional[
    BlockingConnection
]:  # Updated return type to Optional[BlockingConnection]
    """Connect to RabbitMQ and return a blocking connection handle"""

    _rmq_connection = None
    try:
        _rmq_connection = BlockingConnection(
            ConnectionParameters(
                host=conn_config.host,
                port=conn_config.port,
                heartbeat=0,
                credentials=PlainCredentials(
                    username=conn_config.user,
                    password=conn_config.password,
                ),
            )
        )
        assert _rmq_connection is not None

    except (
        pikaExceptions.ChannelWrongStateError,
        pikaExceptions.AMQPConnectionError,
    ) as _exception:
        if raise_exceptions:
            raise _exception

    return _rmq_connection



"""MQP REST API Client"""

from posixpath import join
from typing import Dict, Optional

import requests  # type: ignore
from decouple import config  # type: ignore

from .base_client import BaseClient

MQP_API_VERSION: str = "v1"
REQUEST_TIMEOUT: int = 10


def _fetch_hardcoded_url() -> str:
    """Default URL for MQP REST API"""
    return "https://portal.quantum.lrz.de:4000"


# pylint: disable=too-few-public-methods
class RESTClient(BaseClient):
    """REST API Client Base Class for basic REST functions"""

    def __init__(self, token: str, url: Optional[str] = None) -> None:
        self.token = token

        self.url = url or config("MQP_URL", default=None) or _fetch_hardcoded_url()
        if not self.url:
            raise RuntimeError("No URL provided for MQP.")

    def _headers(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.token}"}

    def get(self, path: str) -> dict:
        """GET request to the specified path"""
        assert isinstance(self.url, str)
        try:
            response = requests.get(
                join(self.url, MQP_API_VERSION, path),
                headers=self._headers(),
                timeout=REQUEST_TIMEOUT,
            )
            response.raise_for_status()
        except requests.exceptions.JSONDecodeError:
            print(f"Error decoding JSON response from {path}")
            return {}
        except requests.exceptions.Timeout:
            print(f"Request to {path} timed out")
            return {}
        except requests.exceptions.RequestException as e:
            print(f"Error fetching data from {path}: {e}")
            return {}
        return response.json()

    def post(self, path: str, data: dict) -> dict:
        """POST request to the specified path"""
        assert isinstance(self.url, str)
        try:
            response = requests.post(
                join(self.url, MQP_API_VERSION, path),
                json=data,
                headers=self._headers(),
                timeout=REQUEST_TIMEOUT,
            )
            response.raise_for_status()
        except requests.exceptions.Timeout:
            print(f"Request to {path} timed out")
            return {}
        except requests.exceptions.RequestException as e:
            print(f"Error posting data to {path}: {e}")
            return {}
        return response.json()

    def delete(self, path: str) -> None:
        """DELETE request to the specified path"""
        assert isinstance(self.url, str)
        try:
            response = requests.delete(
                join(self.url, MQP_API_VERSION, path),
                headers=self._headers(),
                timeout=REQUEST_TIMEOUT,
            )
            response.raise_for_status()
        except requests.exceptions.Timeout:
            print(f"Request to {path} timed out")
        except requests.exceptions.RequestException as e:
            print(f"Error deleting data from {path}: {e}")



"""MQSS Client package"""

from .job import CircuitJobRequest, HamiltonianJobRequest, JobStatus, Result
from .mqss_client import MQSSClient
from .resource_info import ResourceInfo

__all__ = [
    "MQSSClient",
    "ResourceInfo",
    "CircuitJobRequest",
    "HamiltonianJobRequest",
    "JobStatus",
    "Result",
]
